@echo off

:: Introduction is located after local function definitions.
if "%~1"=="" goto :_#intro
if "%~1"=="init" goto :_#init

:: Process calls from user macros.
setlocal enableDelayedExpansion
%@disableErrorMes%
set "_$ed=" & set "rvar=" & set "rval=ERROR"
call %* 2>NUL && (
  if not defined rvar if defined rval %_@showResult% "!rval!"
) || %_@showErrorMesB% "!errorlevel!"
endlocal & (
  if not "%rvar%"=="" set "%rvar%=%rval%"
  set "_$st_date=%_$st_date%" & set "_$st_time=%_$st_time%" & set "_$st_dst=%_$st_dst%"
)
exit /b %errorlevel%

::===================================================================
::---------------- BEGIN LOCAL FUNCTION DEFINITIONS -----------------
:: These are only for internal usage and should never be called directly.

:_#runb
  set "file=%1" & set "var=%2"
  call :_#currentDateTime stDate stTime stDST
  %_@evalRsltVar% %var: =@% rvar
  %_@evalStrayArg% %3
  if "%file%"=="" exit /b %_$EEC_FNSPEC%
  if not exist %file% set "_$ed=%file%" & exit /b %_$EEC_FNFOUND%
  echo Benchmark started - waiting for script or application to finish...
  call %file%
  call :_#currentDateTime spDate spTime spDST
  %_@evalDates% "%stDate%" "%spDate%"
  call :_#calcDuration "%stDate%" "%spDate%" "%stTime%" "%spTime%" "%stDST%" "%spDST%" rval
  echo Benchmark completed.
  exit /b 0

:_#startb
  set "rval="
  %_@evalStrayArg% %1
  call :_#currentDateTime _$st_date _$st_time _$st_dts
  echo Manual benchmark started - waiting for stop call...
  exit /b 0

:_#stopb
  set "var=%1"
  call :_#currentDateTime spDate spTime spDST
  %_@evalRsltVar% %var: =@% rvar
  %_@evalStrayArg% %2
  %_@evalDates% "%_$st_date%" "%spDate%"
  call :_#calcDuration "%_$st_date%" "%spDate%" "%_$st_time%" "%spTime%" "%_$st_dst%" "%spDST%" rval
  set "_$st_date=" & set "_$st_time=" & set "_$st_dst="
  echo Benchmark completed.
  exit /b 0

:_#setResultsToFormattedTime
  %_@evalStrayArg% %1
  set "rval=E" & set "rvar=_$formatted_time"
  exit /b 0

:_#setResultsToTotalTime
  %_@evalStrayArg% %1
  set "rval=" & set "rvar=_$formatted_time"
  exit /b 0

:: {date(yy:MM:dd) %1} {time(HH:mm:ss.ms) %2} {dst(BOOL:TRUE,FALSE) %3}
:_#currentDateTime
  for /f %%a in ('wmic os get LocalDateTime ^| find " "') do set "dts=%%a"
  for /f %%a in ('wmic computersystem get DaylightInEffect ^| find " "') do set "dst=%%a"
  set "%1=%dts:~0,4%/%dts:~4,2%/%dts:~6,2%"
  set "%2=%dts:~8,2%:%dts:~10,2%:%dts:~12,2%.%dts:~15,3%"
  set "%3=!dst!"
  exit /b 0

:: [startDate(yy:MM:dd) %1] [stopDate(yy:MM:dd) %2] [startTime(HH:mm:ss.ms) %3] [stopTime(HH:mm:ss.ms) %4]
:: [startDST(BOOL:TRUE,FALSE) %5] [stopDST(BOOL:TRUE,FALSE) %6] {result(std_spd_stt_spt_d_h_m_s_ms) %7}
:_#calcDuration
  %_@extractDate% "%~1" stYear stMonth stDay
  %_@extractDate% "%~2" spYear spMonth spDay
  %_@daysBetweenDates% "%stYear%" "%spYear%" "%stMonth%" "%spMonth%" "%stDay%" "%spDay%" days
  %@multi% "%days%" "%_$DAY_MS%" dateMS
  %_@convertTimeToMS% "%~3" startMS
  %_@convertTimeToMS% "%~4" stopMS
  %@subi% "%stopMS%" "%startMS%" timeMS
  %_@applyDST% "%~5" "%~6" "%timeMS%" dstTimeMS
  %@addi% "%dateMS%" "%dstTimeMS%" ms
  if defined _$formatted_time (
    %@divi% "%ms%" "%_$DAY_MS%" day
    %@modi% "%ms%" "%_$DAY_MS%" dayMod
    %@divi% "!dayMod!" "%_$HOUR_MS%" hour
    %@modi% "!dayMod!" "%_$HOUR_MS%" hourMod
    %@divi% "!hourMod!" "%_$MIN_MS%" min
    %@modi% "!hourMod!" "%_$MIN_MS%" minMod
    %@divi% "!minMod!" "%_$SEC_MS%" sec
    %@modi% "!minMod!" "%_$SEC_MS%" secMod
    %@modi% "!secMod!" "%_$SEC_MS%" ms
  ) else (
    %@setDeciPlaces% "6"
    %@setRoundDigit% "5"
    %@divf% "%ms%" "%_$DAY_MS%" day
    %@divf% "%ms%" "%_$HOUR_MS%" hour
    %@divf% "%ms%" "%_$MIN_MS%" min
    %@divf% "%ms%" "%_$SEC_MS%" sec
  )
  set "%~7=%~1_%~2_%~3_%~4_%day%_%hour%_%min%_%sec%_%ms%"
  exit /b 0

::---------------- END LOCAL FUNCTION DEFINITIONS -----------------
::=================================================================

::=====================================================
::---------------- BEGIN INTRODUCTION -----------------

:_#intro
echo  ------------------------------------------------------------------------------
echo  benchmark.bat v1.1
echo  Author: Ronnie Fernandez
echo  Email: RonDeCoder@tutamail.com
echo.
echo  v1.0 (2018-1-12)
echo  * Initial release.
echo  v1.1 (2018-1-24)
echo  * Fixed bug where user macros can't be called after changing directory.
echo  * Script will now run even if arithmetic.bat is initialized from a different
echo    location.
echo.
echo  This script was created for benchmarking other scripts or applications.
echo  It is region independant and will keep track of leap years and daylight
echo  savings time where applicable. With current date and time standards, it
echo  should correctly produce results for benchmarks lasting over 300000 years
echo  but that is, of course, untested and therefore NOT guaranteed!
echo.
echo  This script requires arithmetic.bat to process larger integers and
echo  floating-point numbers. In addition, in order to reliably retrieve the date
echo  and time from the user's machine, WMIC must be available for this script to
echo  work. It is native on systems starting with Windows XP Professional and up.
echo.
echo  To use this script, simply call this file by inserting the line
echo  "call benchmark.bat init" at the beginning of your batch file (before setting
echo  delayed expansion) or type "benchmark init" if using from CMD. You will then
echo  be able to utilize all of the following user macros from then on.
echo.
echo  ---------- USER MACROS ----------
echo.
echo  @runb - Runs a benchmark on a script or application
echo  arg1: [file:(filename)]
echo  arg2: {result(std_spd_stt_spt_d_h_m_s_ms)}
echo.
echo  @startb - Starts a manual benchmark
echo.
echo  @stopb - Stops a manual benchmark and displays/saves result
echo  arg1: {result(std_spd_stt_spt_d_h_m_s_ms)}
echo.
echo  @setResultsToFormattedTime - Sets the result output to formatted time
echo.
echo  @setResultsToTotalTime - Sets the result output to total time
echo.
echo  NOTE: If no 'result' variable is specified, the result will be displayed
echo  automatically.
echo.
echo  ------------------------------------------------------------------------------
pause
exit /b 0

::---------------- END INTRODUCTION -----------------
::===================================================

::=======================================================
::---------------- BEGIN INITIALIZATION -----------------
:: This part will load all the variables and macros. This will only happen once.

:_#init

:: Exit if already initialized
if "%_$init_benchmark%"=="TRUE" exit /b 0

:: Exit if delayed expansion is currently enabled.
if "!expansion!"=="" (
  echo ERROR {i}:
  echo Delayed expansion cannot be enabled before initialization.
  exit /b 1
)
:: Check if WMIC is present
WMIC.EXE Alias /? >NUL 2>&1 || (
  echo ERROR {i}:
  echo WMIC.EXE not found. This script requires Windows XP Pro or higher.
  exit /b 1
)
:: Check if arithmetic.bat is already initialized or present in the same directory.
if not "%_$init_arithmetic%"=="TRUE" (  
  if not exist arithmetic.bat (
    echo ERROR{i}:
    echo The required file 'arithmetic.bat' was not found.
    exit /b 1
  )
  call arithmetic.bat init
)

::==================================================================
::---------------- BEGIN LOCAL VARIABLE DEFINITIONS ----------------
:: The following variables are for internal usage and should never be overridden.

:: Signal completion of initialization
set "_$init_benchmark=TRUE"

:: Define evaluation error codes
set "_$EEC_UISDT=4"
set "_$EEC_SDTGTR=5"
set "_$EEC_EMYRS=6"
set "_$EEC_FNSPEC=7"
set "_$EEC_FNFOUND=8"

:: Define values for date and time variables
set "_$DAY_MS=86400000"
set "_$HOUR_MS=3600000"
set "_$MIN_MS=60000"
set "_$SEC_MS=1000"
set "_$DAYS_IN_YEAR=366"
set "_$MAX_YEARS=316800"

:: Define a array with the number of days in each month
set "_$DAYS_INSIDE[1]=31"
set "_$DAYS_INSIDE[2]=29"
set "_$DAYS_INSIDE[3]=31"
set "_$DAYS_INSIDE[4]=30"
set "_$DAYS_INSIDE[5]=31"
set "_$DAYS_INSIDE[6]=30"
set "_$DAYS_INSIDE[7]=31"
set "_$DAYS_INSIDE[8]=31"
set "_$DAYS_INSIDE[9]=30"
set "_$DAYS_INSIDE[10]=31"
set "_$DAYS_INSIDE[11]=30"
set "_$DAYS_INSIDE[12]=31"

:: Define a array with the number of days between months
for /l %%a in (1,1,12) do (
  set "total=0"
  for /l %%b in (%%a,1,12) do (
	setlocal enableDelayedExpansion
    for /f "tokens=1-2" %%c  in ("!total! !_$DAYS_INSIDE[%%b]!") do (
	  endlocal & set "_$DAYS_BETWEEN[%%a.%%b]=%%c" & set /a "total=%%c+%%d"
	)
  )
  set "total=0"
  for /l %%b in (%%a,-1,1) do if %%b neq %%a (
	setlocal enableDelayedExpansion
	for /f "tokens=1-2" %%c  in ("!_$DAYS_BETWEEN[1.%%b]! !_$DAYS_BETWEEN[%%a.12]!") do (
	  endlocal & set /a "_$DAYS_BETWEEN[%%a.%%b]=%%c+%%d+31"
	)
  )
)

:: Store bias for daylight savings time in milliseconds
for /f %%a in ('wmic timezone get DaylightBias ^| find " "') do set /a "_$DTS_BIAS=%%a*_$MIN_MS"

:: Set default values
set "_$st_date="
set "_$st_time="
set "_$st_dst="
set "_$formatted_time=E"

::---------------- END LOCAL VARIABLE DEFINITIONS ----------------
::================================================================

::==============================================================
::---------------- BEGIN USER MACRO DEFINITIONS ----------------
:: These are the only macros intended to be utilized by the user and
:: should never be overridden.

set @runb=call "%~f0" :_#runb,
set @startb=call "%~f0" :_#startb,
set @stopb=call "%~f0" :_#stopb,
set @setResultsToFormattedTime=call "%~f0" :_#setResultsToFormattedTime,
set @setResultsToTotalTime=call "%~f0" :_#setResultsToTotalTime,

::---------------- END USER MACRO DEFINITIONS ----------------
::============================================================

::===============================================================
::---------------- BEGIN LOCAL MACRO DEFINITIONS ----------------
:: These are only for internal usage and should never be overridden.

:: [startDate(yy:MM:dd) %1] [stopDate(yy:MM:dd) %2]
set _@evalDates=%_$BEGIN_MACRO% %\n%
  if "%%~1"=="" exit /b %_$EEC_UISDT% %\n%
  set "d=%%~1/%%~2" %\n%
  for /f "tokens=1-6 delims=/ " %%a in ("!d!") do ( %\n%
    if %%a gtr %_$MAX_YEARS% set "_$ed={START DATE} %%~1" ^& exit /b %_$EEC_EMYRS% %\n%
    if %%d gtr %_$MAX_YEARS% set "_$ed={STOP DATE} %%~2" ^& exit /b %_$EEC_EMYRS% %\n%
    if %%a gtr %%d set "_$ed={START DATE} %%~1 {STOP DATE} %%~2" ^& exit /b %_$EEC_SDTGTR% %\n%
    if %%a equ %%d ( %\n%
	  if %%b gtr %%e set "_$ed={START DATE} %%~1 {STOP DATE} %%~2" ^& exit /b %_$EEC_SDTGTR% %\n%
      if %%b equ %%e if %%c gtr %%f set "_$ed={START DATE} %%~1 {STOP DATE} %%~2" ^& exit /b %_$EEC_SDTGTR% %\n%
    ) %\n%
  ) %\n%
%_$END_MACRO%

:: [date(yy:MM:dd) %1] {year(uint) %2} {month(uint) %3} {day(uint) %4}
set _@extractDate=%_$BEGIN_MACRO% %\n%
  for /f "Tokens=1-3 delims=/ " %%a in ("%%~1") do ( %\n%
	set "%%~2=%%a" ^& set "m=%%b" ^& set "d=%%c" %\n%
    if "!m:~0,1!"=="0" (set "%%~3=!m:~1,1!") else (set "%%~3=!m!") %\n%
	if "!d:~0,1!"=="0" (set "%%~4=!d:~1,1!") else (set "%%~4=!d!") %\n%
  ) %\n%
%_$END_MACRO%

:: [startDST(BOOL:TRUE,FALSE) %1] [stopDST(BOOL:TRUE,FALSE) %2] [milliseconds(uint) %3] {dstMS(uint) %4}
set _@applyDST=%_$BEGIN_MACRO% %\n%
  set "%%~4=%%~3" %\n%
  if /i "%%~1"=="TRUE" ( %\n%
    if "%%~2"=="FALSE" set /a "%%~4-=!_$DTS_BIAS!" %\n%
  ) else if /i "%%~1"=="FALSE" if "%%~2"=="TRUE" set /a "%%~4+=!_$DTS_BIAS!" %\n%
%_$END_MACRO%

:: [time(HH:mm:ss.#ms#) %1] {milliseconds(HH:mm:ss.ms) %2}
set _@convertTimeToMS=%_$BEGIN_MACRO% %\n%
  for /f "tokens=1-4 delims=:. " %%a in ("%%~1") do ( %\n%
    set "%%~2=0" %\n%
    for /f "tokens=* delims=0" %%P in ("%%a") do if not "%%P"=="" set /a "%%~2=%%P*%_$HOUR_MS%" %\n%
	for /f "tokens=* delims=0" %%P in ("%%b") do if not "%%P"=="" set /a "%%~2+=%%P*%_$MIN_MS%" %\n%
	for /f "tokens=* delims=0" %%P in ("%%c") do if not "%%P"=="" set /a "%%~2+=%%P*%_$SEC_MS%" %\n%
	for /f "tokens=* delims=0" %%P in ("%%d") do if not "%%P"=="" set /a "%%~2+=%%P" %\n%
  ) %\n%
%_$END_MACRO%

:: [startYear(uint) %1] [stopYear(uint) %2] [startMonth(uint) %3] [stopMonth(uint) %4]
:: [startDay(uint) %5] [stopDay(uint) %6] {result(uint) %7}
set _@daysBetweenDates=%_$BEGIN_MACRO% %\n%
  if %%~4 lss %%~3 (set /a "yrs=%%~2-%%~1-1") else (set /a "yrs=%%~2-%%~1") %\n%
  set /a "%%~7=(!_$DAYS_BETWEEN[%%~3.%%~4]!+(%%~6-%%~5))+(!yrs!*%_$DAYS_IN_YEAR%)" %\n%
  set /a "%%~7-=((%%~2-%%~1)+1)" %\n%
  if %%~3 geq 3 set /a "%%~7+=1" %\n%
  if %%~4 lss 3 set /a "%%~7+=1" %\n%
  for %%P in (-100 +4 +400) do ( %\n%
    set "dr=%%P" ^& set "sn=!dr:~0,1!"  ^& set "dr=!dr:~1!" %\n%
    set /a "rA=%%~1 %% !dr!" ^& set /a "rB=%%~2 %% !dr!" %\n%
    if !rA! equ 0 (set "nl=%%~1") else (set /a "nl=!dr!-!rA!+%%~1") %\n%
    if !rB! equ 0 (set "pl=%%~2") else (set /a "pl=%%~2-!rB!") %\n%
    if !pl! geq !nl! ( %\n%
      set /a "n=((!pl!-!nl!)/!dr!)+1" %\n%
      if %%~1 equ !nl! if %%~3 geq 3 set /a "n-=1" %\n%
      if %%~2 equ !pl! if %%~4 lss 3 set /a "n-=1" %\n%
	  set /a "%%~7!sn!=!n!" %\n%
    ) %\n%
  ) %\n%
%_$END_MACRO%

:: [result(std_spd_stt_spt_d_h_m_s_ms) %1]
set _@showResult=%_$BEGIN_MACRO% %\n%
  for /f "tokens=1-9 delims=_ " %%a in ("%%~1") do ( %\n%
    echo. %\n%
    echo --------------------------------------- %\n%
    echo ########## Benchmark Results ########## %\n%
    echo --------------------------------------- %\n%
    echo Started     : %%a @ %%c %\n%
    echo Ended       : %%b @ %%d %\n%
    echo --------------------------------------- %\n%
    echo Days:       : %%e %\n%
    echo Hours:      : %%f %\n%
    echo Minutes     : %%g %\n%
    echo Seconds     : %%h %\n%
    echo Milliseconds: %%i %\n%
    echo --------------------------------------- %\n%
    echo. %\n%
    pause %\n%
  ) %\n%
%_$END_MACRO%

:: [errorcode(uint) %1]
set _@showErrorMesB=%_$BEGIN_MACRO% %\n%
  if %%~1 gtr 0 ( %\n%
    echo ERROR {%%~1}: %\n%
    if %%~1 equ %_$EEC_ICOMM% ( %\n%
	  echo Invalid command. %\n%
    ) else if %%~1 equ %_$EEC_IRSLT% ( %\n%
	  echo The result argument cannot contain spaces, '@', or begin with '_$' %\n%
	) else if %%~1 equ %_$EEC_TMA% ( %\n%
	  echo Too many arguments specified: %\n%
	) else if %%~1 equ %_$EEC_UISDT% ( %\n%
	  echo The start date and time was uninitialized. Use @startb macro first. %\n%
	) else if %%~1 equ %_$EEC_SDTGTR% ( %\n%
      echo The start date was more advanced than the stop date: %\n%
	) else if %%~1 equ %_$EEC_EMYRS% ( %\n%
      echo The date exceeded the maximum number of years {%_$MAX_YEARS%}: %\n%
	) else if %%~1 equ %_$EEC_FNSPEC% ( %\n%
      echo The filename for the script or application was not specified. %\n%
	) else if %%~1 equ %_$EEC_FNFOUND% ( %\n%
      echo The specified file was not found: %\n%
	) else echo Unknown %\n%
	if defined _$ed echo !_$ed! %\n%
  ) %\n%
%_$END_MACRO%

::---------------- END LOCAL MACRO DEFINITIONS ----------------
::=============================================================

::---------------- END INITIALIZATION ----------------
::====================================================